library('fda')
library('Matrix')

runnames = c('Cycle 1','Cycle 2','Cycle 3','Cycle 4')

# Load in data

truck = read.table('1truck.dat',head=TRUE)

Y = list()
X = list()

# Which cycles to use and create functional data objects  to obtain accelerations
which = c(4,6,2,5)
for(j in 1:length(which)){

  i = which[j]
  Y[[j]] = truck[truck[,3]==i,2]
  X[[j]] = matrix(truck[truck[,3]==i,1],sum(truck[,3]==i),1)
  
  rr = c(1,length(X[[j]]))
  bb = create.bspline.basis(rr,norder=6,breaks=1:rr[2])

  fd0 = smooth.basis(1:rr[2],X[[j]],fdPar(bb,3,0.1)) 
#  plotfit.fd(X[[i]],1:rr[2],fd0$fd)
  
  v = eval.fd(1:rr[2],fd0$fd,1)
#  x = eval.fd(1:rr[2],fd0$fd,0)
#  plot(v,type='l')

  X[[j]] = cbind(X[[j]],v)
}




print('X')

# Now for each length l create a matrix that looks at the previous l seconds of data

lt = 20
lmax = 40
lmin = 5



ncyc = 4
n = rep(0,ncyc)

Xmat = list(len=ncyc)
YY = c()

nremove = c(9,19,18,9)

for(i in 1:ncyc){
  n[i] = nrow(X[[i]])
  Xmat[[i]] = list(len=ncol(X[[i]]))
  for(j in 1:ncol(X[[i]])){
    Xmat[[i]][[j]] =  matrix(X[[i]][,j],nrow(X[[i]])+1,(lmax+1),byrow=FALSE)[1:(n[i]-lmax-1),rev(1:(lmax+1))]
    Xmat[[i]][[j]] = Xmat[[i]][[j]][-(1:nremove[i]),]
  }
  n[i] = nrow(Xmat[[i]][[1]])
  YY = c(YY,Y[[i]][-(1:(nremove[i]+lmax+1))])
}

YY = YY*100
N = cumsum(c(1,n))

lwds=c(2.5,2.5,2,2)
ltys = c(1,3,1,2)
cols = colors()[c(213,153,153,153)]

# Plot the data
#postscript('../ConvFigures/Data.eps')
par(mfrow=c(3,1),mar=c(5,5,0,1))
plot(YY[N[1]:(N[2]-1)],type='l',ylab='log PM',xlab='',xlim=c(0,1200),cex.lab=1.5,cex.axis=1.5,lwd=lwds[1],lty=ltys[1],col=cols[1])
for(j in 2:length(n)){ lines(YY[N[j]:(N[j+1]-1)],lty=ltys[j],lwd=lwds[j],col=cols[j]) }
plot(Xmat[[1]][[1]][,1],xlim=c(0,1200),type='l',ylab='Speed',xlab='',ylim=c(0,70),
   cex.lab=1.5,cex.axis=1.5,lwd=lwds[1],lty=ltys[1],col=cols[1])
for(j in 2:length(n)){ lines(Xmat[[j]][[1]][,1],lty=ltys[j],lwd=lwds[j],col=cols[j]) }
legend('topright',legend=runnames,lwd=lwds,lty=ltys,col=cols,cex=1.5)
plot(Xmat[[1]][[2]][,1],xlim=c(0,1200),type='l',ylab='Acceleration',xlab='Seconds',ylim=c(-5,3),cex.lab=1.5,cex.axis=1.5,lwd=lwds[1],lty=ltys[1],col=cols[1])
for(j in 2:length(n)){ lines(Xmat[[j]][[2]][,1],lty=ltys[j],lwd=lwds[j],col=cols[j]) }
#dev.off()



print('Xmat')

# Now put this into bases to combine with covariates. 

basislist = vector('list',len=lmax-lmin)
basismatlist = vector('list',len=lmax-lmin)
Xmatlist = vector('list',len=lmax-lmin)
Penmatlist =  vector('list',len=lmax-lmin)
b = rep(0,lmax-lmin)


for(i in 1:(lmax-lmin)){
  basislist[[i]] = create.bspline.basis(c(0,i+lmin),norder=4,breaks = 0:(i+lmin))
  b[i] = basislist[[i]]$nbasis

  tPen = eval.penalty(basislist[[i]],int2Lfd(2))
  Penmatlist[[i]] =  rbind(matrix(0,4,4+2*(ncol(tPen)-1)),cbind(matrix(0,2*(nrow(tPen)-1),4),diag(2)%x%tPen[-b[i],-b[i]]))
#  Penmatlist[[i]] =  rbind(rep(0,1+1*(ncol(tPen)-1)),cbind(rep(0,1*(nrow(tPen)-1)),diag(1)%x%tPen[-b[i],-b[i]]))
  basismatlist[[i]] = eval.basis(0:(i+lmin),basislist[[i]])[,-b[i]]

  for(j in 1:ncyc){
    tX = matrix(0,n[j],ncyc)
    tX[,j] = 1
    for(k in 1:ncol(X[[j]])){
#      tX = cbind(tX,Xmat[[j]][[k]][,lmax+1-(0:(i+lmin))]%*%basismatlist[[i]])
      tX = cbind(tX,Xmat[[j]][[k]][,1+(0:(i+lmin))]%*%basismatlist[[i]])
    }
    Xmatlist[[i]] = rbind(Xmatlist[[i]],tX)
  }
}

print('Xmatlist')

# Hat matrices for each length and lambda

lambda = 10^seq(-3,2,0.5)

hatmat = vector('list',len=length(lambda))

df = matrix(0,length(lambda),lmax-lmin)

for(i in 1:length(lambda)){
   hatmat[[i]] = vector('list',len=lmax-lmin)
   for(j in 1:(lmax-lmin)){
     hatmat[[i]][[j]] = solve(t(Xmatlist[[j]])%*%Xmatlist[[j]]+lambda[i]*Penmatlist[[j]],t(Xmatlist[[j]]))
     df[i,j] = sum(diag(Xmatlist[[j]]%*%hatmat[[i]][[j]]))
   }
}


print('hatmat')
# And estimate including various cross validation measures. 


OCV = matrix(0,length(lambda),lmax-lmin)
OCV2 = matrix(0,length(lambda),lmax-lmin)
coef = vector('list',len=length(lambda))

CVpreds = vector('list',len=4)
for(i in 1:4){ CVpreds[[i]] = array(0,c(N[i+1]-N[i],length(lambda),lmax-lmin)) }

for(i in 1:length(lambda)){
   coef[[i]] = vector('list',len=lmax-lmin)
   for(j in 1:(lmax-lmin)){
      coef[[i]][[j]] = hatmat[[i]][[j]]%*%YY
      pred = Xmatlist[[j]]%*%coef[[i]][[j]]
      OCV[i,j] = apply( ((YY-pred)/(length(YY)-df[i,j]))^2,2,sum)


      for(k in 1:(length(N)-1)){
        I = (N[k]:(N[k+1]-1))
        tX = Xmatlist[[j]][-I,-k]
        pred =  Xmatlist[[j]][I,-k]%*%solve(t(tX)%*%tX+lambda[i]*Penmatlist[[j]][-k,-k],t(tX)%*%YY[-I])
        CVpreds[[k]][,i,j] = pred - mean(pred) + mean(YY[I])
        OCV2[i,j] = OCV2[i,j] + apply( (pred-mean(pred)-YY[I]+mean(YY[I]))^2,2,sum)
     }
  }
}

print('results')

# And find the best estimates

best = which.min(OCV)
bestj = floor((best-1)/length(lambda))
besti = best - length(lambda)*bestj
bestj = bestj + 1

par(mfrow=c(2,1))
plot( fd( c(coef[[besti]][[bestj]][ncyc+(1:(basislist[[bestj]]$nbasis-1))],0), basislist[[bestj]]))
plot( fd( c(coef[[besti]][[bestj]][basislist[[bestj]]$nbasis+ncyc-1+(1:(basislist[[bestj]]$nbasis-1))],0), basislist[[bestj]]))


# OK, now let's try a bootstrap
nboot = 500
I = matrix(1:3140,20,157)
K = 20
M = 157

zerotemplate = ( matrix(1:41,41,40,byrow=FALSE) <= matrix(1:40,41,40,byrow=TRUE))

nj = ncol(basismatlist[[bestj]])
pred =  Xmatlist[[bestj]]%*%coef[[besti]][[bestj]]      
resid = YY - pred

isse = rep(0,ncyc)
isse2 = rep(0,ncyc)
icor = rep(0,ncyc)

par(mfrow=c(2,2))
for(i in 1:ncyc){
  ylims = range(c(pred[N[i]:(N[i+1]-1)], YY[N[i]:(N[i+1]-1)]))
  isse[i] = round(mean( (YY[N[i]:(N[i+1]-1)] - pred[N[i]:(N[i+1]-1)])^2 ),digits=4)
  isse2[i] = round(mean( (YY[N[i]:(N[i+1]-1)] - CVpreds[[i]][,besti,bestj])^2 ),digits=4)
  icor[i] = round(cor(YY[N[i]:(N[i+1]-1)], pred[N[i]:(N[i+1]-1)]),digits=4)
  plot(pred[N[i]:(N[i+1]-1)],type='l',lwd=2,xlab='Time',ylab='PM Values',,ylim=ylims,
    cex.lab=1.5,cex.axis=1.5,cex.main=1.5,main=paste(runnames[i],'    SSE:',isse2[i],'  correlation',icor[i]))
  lines(YY[N[i]:(N[i+1]-1)],lwd=2,lty=2,col=1)
}


# Look at residulas

resmat = matrix(0,N[ncyc+1]-1,40)
for(k in 1:ncyc){
  tempres = resid[N[k]:(N[k+1]-1)]
  tempresmat = matrix(tempres,n[k]+1,40)*rbind( matrix(1,n[k]-40,40),zerotemplate)
  resmat[N[k]:(N[k+1]-1),] = tempresmat[1:n[k],]
}

autocov = (t(resmat)%*%resid)*(1/(N[ncyc+1] - ncyc*(1:40)))

print('autocov')

# Look at autocovariances for each cycle independently

autocovs = matrix(0,40,ncyc)
for(k in 1:ncyc){
   autocovs[,k] = (t(resmat[N[k]:(N[k+1]-1),])%*%resid[N[k]:(N[k+1]-1)])*(1/((N[k+1]-N[k]) - (1:40)))

}

autocors = autocovs%*%diag(1/autocovs[1,])

#postscript('../ConvFigures/autocovariances.eps')
matplot(autocovs,type='b',pch=1:4,lwd=2,cex.lab=1.5,cex.axis=1.5, ylab='autocovariance',xlab='lag',col=1)
lines(autocov,lwd=2) #/autocov[1],lwd=2)
abline(h=0)
legend('topright',c('Average',runnames),pch=c(NA,1:4),lwd=2,cex=1.5,lty=1:4,col=1)
#dev.off()

  # Here we go for variances of y and of the blocked version
  
  # First define a transformation C from block bootstrap back to having the correct
  # covariance. 

  M2 = bandSparse(20,k=0:19,diag=matrix(autocov[1:20],20,20,byrow=TRUE),symm=TRUE)
  eM2 = eigen(M2);
  dM2 = 1/sqrt(eM2$val)
  dM2[eM2$val<0] = 0
  eM2.2 = eM2$vec%*%diag(dM2)%*%t(eM2$vec)

  CC = list(ncyc)
  M1 = list(ncyc)
  for(k in 1:ncyc){
    M1[[k]] = as.matrix(bandSparse(n[k],k=0:19,diag=matrix(autocov[1:20],n[k],20,byrow=TRUE),symm=TRUE))
    eM1 = eigen(M1[[k]])    
    CC[[k]] = eM1$vec%*%(t(eM1$vec)*sqrt(eM1$val*(eM1$val>0)))%*%( diag(n[k]/K)%x%eM2.2 )
  }
 
  C = bdiag(CC)  

print('C')
  

  cmat = hatmat[[besti]][[bestj]]%*%bdiag(M1)%*%t(hatmat[[besti]][[bestj]])
  sdnoboot = sqrt(diag(basismatlist[[bestj]]%*%cmat[ncyc+(1:nj),ncyc+(1:nj)]%*%t(basismatlist[[bestj]])) )
  sdnoboot = c(sdnoboot,rep(0,lmax+1-length(sdnoboot)))
  sdnoboot2 = sqrt(diag(basismatlist[[bestj]]%*%cmat[ncyc+nj+(1:nj),ncyc+nj+(1:nj)]%*%t(basismatlist[[bestj]])) )
  sdnoboot2 = c(sdnoboot2,rep(0,lmax+1-length(sdnoboot2)))


 # Generate matrix of bootstrapped blocks 

  Rmat = matrix(resid[as.vector(I[,sample(M,M*nboot,replace=TRUE)])],length(resid),nboot)
  Rmat2 = C%*%Rmat

  # Choose and example to plot

#postscript('../ConvFigurs/ResidualTransform2.eps',paper='special',width=10,height=6)
#par(pin=c(8,4))
plot(Rmat2[1:100,10],type='l',lwd=2,#ylim=c(-1.6,1.1),cex.lab=1.5,cex.axis=1.5,
  xlab='time',ylab='residual')
for(i in 1:10){ 
  lines(1:10+10*(i-1),Rmat[1:10 + 10*(i-1),10],lty=2,lwd=2,col=2)
}
legend('bottomright',c('Block Residuals','Transformed Residuals'),lty=c(2,1),col=c(2,1),lwd=2,cex=1.5)
#dev.off()

## Calculate implied autocorrelations with and without transformation

autocovb = matrix(0,21,nboot)
autocovb2 = matrix(0,21,nboot)

for(i in 1:21){
  j = i-1
  for(k in 1:ncyc){
    autocovb[i,] = autocovb[i,] + apply(Rmat[N[k]:(N[k+1]-40),]*Rmat[j+(N[k]:(N[k+1]-40)),],2,mean)
    autocovb2[i,] = autocovb2[i,] + apply(Rmat2[N[k]:(N[k+1]-40),]*Rmat2[j+(N[k]:(N[k+1]-40)),],2,mean)
  }
}

autocovb = autocovb/ncyc
autocovb2 = autocovb2/ncyc


#postscript('../ConvFigures/BlockBootCov2.eps')
matplot(0:20,autocovb,type='l',col='grey',lty=1,xlab='lag',ylab='autocovariance',
  main='Block Bootstrap Covariance',cex.lab=1.5,cex.axis=1.5,cex.main=2)
lines(0:20,autocov[0:21],lwd=2)
#dev.off()

#postscript('../ConvFigures/TransBootCov2.eps')
matplot(0:20,autocovb2,type='l',col='grey',lty=1,xlab='lag',ylab='autocovariance',
  main='Transformed Bootstrap Covariance',cex.lab=1.5,cex.axis=1.5,cex.main=2)
lines(0:20,autocov[0:21],lwd=2)
#dev.off()

  # Now create a matrix of responses using block residual bootstraps
  
  tY = pred %x% matrix(1,1,nboot) + Rmat2
  ttY = pred %x% matrix(1,1,nboot) + Rmat


# Bootstrap everything. 
print('bootstrap')

  tcoef = vector('list',len=length(lambda))
  ttcoef = vector('list',len=length(lambda))
  tOCV = array(0,c(length(lambda),lmax-lmin,ncol(tY)))
  ttOCV = array(0,c(length(lambda),lmax-lmin,ncol(tY)))
  
  for(ii in 1:length(lambda)){
    tcoef[[ii]] = vector('list',len=lmax-lmin)
    ttcoef[[ii]] = vector('list',len=lmax-lmin)
    for(j in 1:(lmax-lmin)){
      hat = diag(Xmatlist[[j]]%*%hatmat[[ii]][[j]])    
    
      tcoef[[ii]][[j]] = hatmat[[ii]][[j]]%*%tY
      tpred = Xmatlist[[j]]%*%tcoef[[ii]][[j]]
      tOCV[ii,j,] = apply( ((tY-tpred)/(1-hat))^2,2,sum) 
      
      ttcoef[[ii]][[j]] = hatmat[[ii]][[j]]%*%ttY 
      ttpred = Xmatlist[[j]]%*%ttcoef[[ii]][[j]]
      ttOCV[ii,j,] = apply( ((ttY-ttpred)/(1-hat))^2,2,sum)   
    
#      for(k in 1:(length(N)-1)){
#        tX = Xmatlist[[j]][-(N[k]:(N[k+1]-1)),-k]
#        tpred =  Xmatlist[[j]][N[k]:(N[k+1]-1),]%*%solve(t(tX)%*%tX+lambda[ii]*Penmatlist[[j]][-k,-k],t(tX)%*%tY[-(N[k]:(N[k+1]-1)),])
#        tOCV[ii,j,] = tOCV[ii,j,] + apply( (tpred-tY[N[k]:(N[k+1]-1),])^2,2,sum)
#        
#        ttpred =  Xmatlist[[j]][N[k]:(N[k+1]-1),]%*%solve(t(tX)%*%tX+lambda[ii]*Penmatlist[[j]][-k,-k],t(tX)%*%ttY[-(N[k]:(N[k+1]-1)),])
#        ttOCV[ii,j,] = ttOCV[ii,j,] + apply( (ttpred-ttY[N[k]:(N[k+1]-1),])^2,2,sum)        
#      }
    }
  }

print('bootselect')

  tbest = apply(tOCV,3,which.min)
  tbestj = floor((tbest-1)/length(lambda))
  tbesti = tbest - length(lambda)*tbestj
  tbestj = tbestj + 1
  
  ttbest = apply(ttOCV,3,which.min)
  ttbestj = floor((ttbest-1)/length(lambda))
  ttbesti = ttbest - length(lambda)*ttbestj
  ttbestj = ttbestj + 1  
    
  selection = cbind(tbesti,tbestj) #,ttbesti,ttbestj) 
  

print('estimates')

  tbhatvals =  matrix(0,(lmax+1),nboot)
  ttbhatvals =  matrix(0,(lmax+1),nboot)
  tbhatvals2 =  matrix(0,(lmax+1),nboot)
  ttbhatvals2 =  matrix(0,(lmax+1),nboot)
  
  tpredvals = matrix(0,length(pred),nboot)

  for(j in 1:nboot){
    njj = ncol(basismatlist[[tbestj[j]]])
    tnjj = ncol(basismatlist[[ttbestj[j]]])
    tbhatvals[1:(tbestj[j]+lmin+1),j] = basismatlist[[tbestj[j]]]%*%tcoef[[tbesti[j]]][[tbestj[j]]][ncyc+(1:njj),j]
    ttbhatvals[1:(ttbestj[j]+lmin+1),j] = basismatlist[[ttbestj[j]]]%*%ttcoef[[ttbesti[j]]][[ttbestj[j]]][ncyc+(1:tnjj),j]
    tbhatvals2[1:(tbestj[j]+lmin+1),j] = basismatlist[[tbestj[j]]]%*%tcoef[[tbesti[j]]][[tbestj[j]]][ncyc+njj+(1:njj),j]
    ttbhatvals2[1:(ttbestj[j]+lmin+1),j] = basismatlist[[ttbestj[j]]]%*%ttcoef[[ttbesti[j]]][[ttbestj[j]]][ncyc+tnjj+(1:tnjj),j]
    
    tpredvals[,j] = Xmatlist[[tbestj[j]]]%*%tcoef[[tbesti[j]]][[tbestj[j]]][,j]
  }  
    

print('analysis')

  sdboot = apply(tbhatvals,1,sd)
  sdboot.2 = apply(ttbhatvals,1,sd)

  quantboot = apply(tbhatvals,1,quantile,c(0.025,0.975))
  quantboot.2 = apply(ttbhatvals,1,quantile,c(0.025,0.975))

  sdboot2 = apply(tbhatvals2,1,sd)
  sdboot2.2 = apply(ttbhatvals2,1,sd)

  quantboot2 = apply(tbhatvals2,1,quantile,c(0.025,0.975))
  quantboot2.2 = apply(ttbhatvals2,1,quantile,c(0.025,0.975))
  
  predboot = apply(tpredvals,1,sd)
  
bhatvals = basismatlist[[bestj]]%*%coef[[besti]][[bestj]][ncyc+(1:nj)]
bhatvals = c(bhatvals,rep(0,lmax+1-length(bhatvals)))

bhatvals2 = basismatlist[[bestj]]%*%coef[[besti]][[bestj]][ncyc+nj+(1:nj)]
bhatvals2 = c(bhatvals2,rep(0,lmax+1-length(bhatvals2)))


#postscript('../ConvFigures/CoefEstimates1.eps')
par(mfrow=c(2,1),mar=c(5,5,1,1))
matplot(bhatvals%x%t(rep(1,3))+sdboot%x%t(c(-2,0,2)),type='l',col=1,lty=c(2,1,2),cex.lab=1.5,cex.axis=1.5,
  ylab = 'beta_velocity',xlab='lag',lwd=2)
matplot(bhatvals%x%t(rep(1,2))+sdnoboot%x%t(c(-2,2)),type='l',col=1,lty=3,add=TRUE,lwd=2)
abline(h=0)
matplot(bhatvals2%x%t(rep(1,3))+sdboot2%x%t(c(-2,0,2)),type='l',col=1,lty=c(2,1,2),cex.lab=1.5,cex.axis=1.5,
  ylab = 'beta_accel',xlab='lag',lwd=2)
matplot(bhatvals2%x%t(rep(1,2))+sdnoboot2%x%t(c(-2,2)),type='l',col=1,lty=3,add=TRUE,lwd=2) 
abline(h=0)
legend('topright',c('Coefficient','Boostrap CI','Delta CI'),lty=1:3,col=1,lwd=2,cex=1.5)
#dev.off()

#postscript('../ConvFigures/Predictions.eps')
par(mfrow=c(2,2),mar=c(3,3,3,1))
for(i in 1:ncyc){
  predmat = matrix(pred[N[i]:(N[i+1]-1)],N[i+1]-N[i],2,byrow=FALSE) + (predboot[N[i]:(N[i+1]-1)]+sqrt(autocov[1]))%*%matrix(c(-2,2),1,2)

  ylims = range(c(predmat, YY[N[i]:(N[i+1]-1)]))
  isse[i] = round(mean( (YY[N[i]:(N[i+1]-1)] - pred[N[i]:(N[i+1]-1)])^2 ),digits=4)
  
  matplot( predmat,type='l',lty=3,xlab='',ylab='',,ylim=ylims,
    cex.lab=1.5,cex.axis=1.5,cex.main=1.5,main=paste(runnames[i],'    SSE:',isse2[i]),lwd=2,col=1)
  
  lines(YY[N[i]:(N[i+1]-1)],lwd=2,lty=1,,col=colors()[213])
  lines(CVpreds[[i]][,besti,bestj],lwd=2,lty=2) 
}
#dev.off()


# Finally, let's look at out of sample concurrent model

concorrs = rep(0,ncyc)
consse = rep(0,ncyc)

for(i in 1:ncyc){
  dd = data.frame(Y = do.call(c,Y[-i]), do.call(rbind,X[-i]))
  pp = predict( lm(Y~.,data=dd),data.frame(X[[i]]) )
  concorrs[i] = cor(pp,Y[[i]])
  consse[i] = mean(  ( Y[[i]] - mean(Y[[i]]) - pp + mean(pp) )^2 )
}

SSEmat = rbind(isse2,100^2*consse)
colnames(SSEmat) = runnames

round(SSEmat,4)

CORmat = rbind(icor,concorrs)
colnames(CORmat) = runnames

round(CORmat,4)

save.image('1Truck_final_02072013.Rdata')