# This file runs a simulation to test the statistical properties
# of functional convolution models. This file should be run after 
# SimulationSetup.R

library('fda')
library('Matrix')

savefile = 'SimAll_estcoefs_4(corrected)_022912.Rdata'


# The following file is created by SimulationSetup.R and contains the design matrices
# for each value of lambda and length of convolution. 

load('TruckRun.Simdata2.Rdata')

# Simulated responses:

Y = as.matrix(read.table('simdata2.txt',head=FALSE))

nsim = ncol(Y)

Y = Y[,1:nsim]


# Now obtain estimates

start = proc.time()
OCV = array(0,c(length(lambda),lmax-lmin,nsim))
coef = vector('list',len=length(lambda))

for(i in 1:length(lambda)){
   coef[[i]] = vector('list',len=lmax-lmin)
   for(j in 1:(lmax-lmin)){
      coef[[i]][[j]] = hatmat[[i]][[j]]%*%Y
      hat = diag(Xmatlist[[j]]%*%hatmat[[i]][[j]])
      pred = Xmatlist[[j]]%*%coef[[i]][[j]]
      OCV[i,j,] = apply( ((Y-pred)/(1-hat))^2,2,sum)
      
#      coef[[i]][[j]] = solve(t(Xmatlist[[j]])%*%Xmatlist[[j]]+lambda[i]*Penmatlist[[j]],t(Xmatlist[[j]])%*%Y)
#   
#      for(k in 1:(length(N)-1)){
#        num = floor( (N[k+1]-N[k])/50 )
#        I = (N[k]:(N[k+1]-1))
#        tX = Xmatlist[[j]][-(N[k]:(N[k+1]-1)),]
#        pred =  Xmatlist[[j]][I,]%*%solve(t(tX)%*%tX+lambda[i]*Penmatlist[[j]],t(tX)%*%Y[-(N[k]:(N[k+1]-1)),])
#        OCV[i,j,] = OCV[i,j,] + apply( (pred-Y[I,])^2,2,sum)
#     }
  }
}


# Work out which is the best length and lambda

best = apply(OCV,3,which.min)
bestj = floor((best-1)/length(lambda))
besti = best - length(lambda)*bestj
bestj = bestj + 1
print(proc.time()-start)


## Plot these

hist(besti,50)
hist(bestj,50)


X11()
par(mfrow=c(2,1))
plot(0:lmax,beta1,type='l',xlim=c(0,lmax),ylim=c(-15,10))
for(i in 1:nsim){
  lines(0:(nrow(basismatlist[[bestj[i]]])-1), basismatlist[[bestj[i]]]%*%coef[[besti[i]]][[bestj[i]]][ncyc+1:ncol(basismatlist[[bestj[i]]]),i])
}
lines(0:lmax,beta1,lwd=2,col=2)


plot(0:lmax,beta2,type='l',xlim=c(0,lmax),ylim=c(-10,100))
for(i in 1:nsim){
  lines(0:(nrow(basismatlist[[bestj[i]]])-1), basismatlist[[bestj[i]]]%*%coef[[besti[i]]][[bestj[i]]][ncyc+ncol(basismatlist[[bestj[i]]])+1:ncol(basismatlist[[bestj[i]]]),i])
}
lines(0:lmax,beta2,lwd=2,col=2)


# Bootstrap

# I, M, K from saved file
            
autocov = matrix(0,40,nboot)
zerotemplate = ( matrix(1:41,41,40,byrow=FALSE) <= matrix(1:40,41,40,byrow=TRUE))

mboot = matrix(0,lmax+1,nsim)
mboot.2 = matrix(0,lmax+1,nsim)
sdboot = matrix(0,lmax+1,nsim)
sdboot.2 = matrix(0,lmax+1,nsim)
sdnoboot = matrix(0,lmax+1,nsim)
quantboot = array(0,c(2,lmax+1,nsim))
quantboot.2 = array(0,c(2,lmax+1,nsim))

mboot2 = matrix(0,lmax+1,nsim)
mboot2.2 = matrix(0,lmax+1,nsim)
sdboot2 = matrix(0,lmax+1,nsim)
sdboot2.2 = matrix(0,lmax+1,nsim)
sdnoboot2 = matrix(0,lmax+1,nsim)
quantboot2 = array(0,c(2,lmax+1,nsim))
quantboot2.2 = array(0,c(2,lmax+1,nsim))


bhat1 = matrix(0,(lmax+1),nsim)
bhat2 = matrix(0,(lmax+1),nsim)

selection = array(0,c(nboot,4,nsim))

NN = nrow(Y)

de = (0:(NN-1))*NN + 1:NN

for(i in 1:nsim){
   print(i)
   start = proc.time()
   ibest = besti[[i]]
   jbest = bestj[[i]]
   nj = nrow(basismatlist[[jbest]])
   njj = ncol(basismatlist[[jbest]])
  # Obtain residuals
             
  bhat1[1:nj,i] = basismatlist[[jbest]]%*%coef[[ibest]][[jbest]][ncyc+(1:njj),i]
  bhat2[1:nj,i] = basismatlist[[jbest]]%*%coef[[ibest]][[jbest]][ncyc+njj+(1:njj),i]   
 
  pred =  Xmatlist[[jbest]]%*%coef[[ibest]][[jbest]][,i]      
  resid = Y[,i] - pred

# Estimate autocovariance + correction C

  resmat = matrix(0,N[ncyc+1]-1,40)
  for(k in 1:ncyc){
    tempres = resid[N[k]:(N[k+1]-1)]
    tempresmat = matrix(tempres,n[k]+1,40)*rbind( matrix(1,n[k]-40,40),zerotemplate)
    resmat[N[k]:(N[k+1]-1),] = tempresmat[1:n[k],]
  }

  autocov[,i] = (t(resmat)%*%resid)*(1/(N[ncyc+1] - ncyc*(1:40)))

  # Here we go for variances of y and of the blocked version

  M2 = bandSparse(20,k=0:19,diag=matrix(autocov[1:20,i],20,20,byrow=TRUE),symm=TRUE)
  eM2 = eigen(M2);
  eM2.2 = eM2$vec%*%diag(1/sqrt(eM2$val))%*%t(eM2$vec)

  CC = list(ncyc)
  M1 = list(ncyc)
  for(k in 1:ncyc){
    M1[[k]] = as.matrix(bandSparse(n[k],k=0:19,diag=matrix(autocov[1:20,i],n[k],20,byrow=TRUE),symm=TRUE))
    eM1 = eigen(M1[[k]])    
    CC[[k]] = eM1$vec%*%diag(sqrt(eM1$val*(eM1$val>0)))%*%t(eM1$vec)%*%( diag(n[k]/K)%x%eM2.2 )
  }
 
  C = bdiag(CC)  

# Delta method standard errors

  cmat = hatmat[[besti[i]]][[bestj[i]]]%*%bdiag(M1)%*%t(hatmat[[besti[i]]][[bestj[i]]])
  sdnoboot[1:nj,i] = sqrt(diag(basismatlist[[bestj[i]]]%*%cmat[ncyc+(1:njj),ncyc+(1:njj)]%*%t(basismatlist[[bestj[i]]])) )
  sdnoboot2[1:nj,i] = sqrt(diag(basismatlist[[bestj[i]]]%*%cmat[ncyc+njj+(1:njj),ncyc+njj+(1:njj)]%*%t(basismatlist[[bestj[i]]])) )

# Generate matrix of bootstrapped blocks 

  Rmat =  eM2$vec%*%diag(sqrt(eM2$val))%*%t(eM2$vec)%*%matrix(rnorm(length(resid)*nboot),K,nboot*M)
  Rmat = matrix(as.vector(Rmat),length(resid),nboot,byrow=FALSE)
   
  Rmat2 = C%*%Rmat
  
  tY = pred %x% matrix(1,1,nboot) + Rmat2
  ttY = pred %x% matrix(1,1,nboot) + Rmat

  
  tcoef = vector('list',len=length(lambda))
  ttcoef = vector('list',len=length(lambda))
  tOCV = array(0,c(length(lambda),lmax-lmin,ncol(tY)))
  ttOCV = array(0,c(length(lambda),lmax-lmin,ncol(tY)))
  
  for(ii in 1:length(lambda)){
    tcoef[[ii]] = vector('list',len=lmax-lmin)
    ttcoef[[ii]] = vector('list',len=lmax-lmin)
    for(j in 1:(lmax-lmin)){
      hat = (Xmatlist[[j]]%*%hatmat[[ii]][[j]])[de]    
    
      tcoef[[ii]][[j]] = hatmat[[ii]][[j]]%*%tY
      tpred = Xmatlist[[j]]%*%tcoef[[ii]][[j]]
      tOCV[ii,j,] = apply( ((tY-tpred)/(1-hat))^2,2,sum) 
      
      ttcoef[[ii]][[j]] = hatmat[[ii]][[j]]%*%ttY 
      ttpred = Xmatlist[[j]]%*%ttcoef[[ii]][[j]]
      ttOCV[ii,j,] = apply( ((ttY-ttpred)/(1-hat))^2,2,sum)   
    
#      for(k in 1:(length(N)-1)){
#        tX = Xmatlist[[j]][-(N[k]:(N[k+1]-1)),]
#        tpred =  Xmatlist[[j]][N[k]:(N[k+1]-1),]%*%solve(t(tX)%*%tX+lambda[ii]*Penmatlist[[j]],t(tX)%*%tY[-(N[k]:(N[k+1]-1)),])
#        tOCV[ii,j,] = tOCV[ii,j,] + apply( (tpred-tY[N[k]:(N[k+1]-1),])^2,2,sum)
#        
#        ttpred =  Xmatlist[[j]][N[k]:(N[k+1]-1),]%*%solve(t(tX)%*%tX+lambda[ii]*Penmatlist[[j]],t(tX)%*%ttY[-(N[k]:(N[k+1]-1)),])
#        ttOCV[ii,j,] = ttOCV[ii,j,] + apply( (ttpred-ttY[N[k]:(N[k+1]-1),])^2,2,sum)        
#      }
    }
  }
  
  tbest = apply(tOCV,3,which.min)
  tbestj = floor((tbest-1)/length(lambda))
  tbesti = tbest - length(lambda)*tbestj
  tbestj = tbestj + 1
  
  ttbest = apply(ttOCV,3,which.min)
  ttbestj = floor((ttbest-1)/length(lambda))
  ttbesti = ttbest - length(lambda)*ttbestj
  ttbestj = ttbestj + 1  
    
  selection[,1:2,i] = cbind(tbesti,tbestj) #,ttbesti,ttbestj) 
  
# Now get standard errors  
  
  tbhatvals =  matrix(0,(lmax+1),nboot)
  ttbhatvals =  matrix(0,(lmax+1),nboot)
  tbhatvals2 =  matrix(0,(lmax+1),nboot)
  ttbhatvals2 =  matrix(0,(lmax+1),nboot)

  for(j in 1:nboot){
    njj = ncol(basismatlist[[tbestj[j]]])
    tnjj = ncol(basismatlist[[ttbestj[j]]])
    tbhatvals[1:(tbestj[j]+lmin+1),j] = basismatlist[[tbestj[j]]]%*%tcoef[[tbesti[j]]][[tbestj[j]]][ncyc+(1:njj),j]
    ttbhatvals[1:(ttbestj[j]+lmin+1),j] = basismatlist[[ttbestj[j]]]%*%ttcoef[[ttbesti[j]]][[ttbestj[j]]][ncyc+(1:tnjj),j]
    tbhatvals2[1:(tbestj[j]+lmin+1),j] = basismatlist[[tbestj[j]]]%*%tcoef[[tbesti[j]]][[tbestj[j]]][ncyc+njj+(1:njj),j]
    ttbhatvals2[1:(ttbestj[j]+lmin+1),j] = basismatlist[[ttbestj[j]]]%*%ttcoef[[ttbesti[j]]][[ttbestj[j]]][ncyc+tnjj+(1:tnjj),j]
  }  


  mboot[,i] = apply(tbhatvals,1,mean)
  mboot.2[,i] = apply(ttbhatvals,1,mean)
   
  sdboot[,i] = apply(tbhatvals,1,sd)
  sdboot.2[,i] = apply(ttbhatvals,1,sd)

  quantboot[,,i] = apply(tbhatvals,1,quantile,c(0.025,0.975))
  quantboot.2[,,i] = apply(ttbhatvals,1,quantile,c(0.025,0.975))

  mboot2[,i] = apply(tbhatvals2,1,mean)
  mboot2.2[,i] = apply(ttbhatvals2,1,mean)
   
  sdboot2[,i] = apply(tbhatvals2,1,sd)
  sdboot2.2[,i] = apply(ttbhatvals2,1,sd)

  quantboot2[,,i] = apply(tbhatvals2,1,quantile,c(0.025,0.975))
  quantboot2.2[,,i] = apply(ttbhatvals2,1,quantile,c(0.025,0.975))
   
  print(proc.time()-start)
  save.image(savefile)


   rm(tcoef,ttcoef,tOCV,ttOCV,C,Rmat,Rmat2,resmat,tY,ttY)
   
}

truthmat1 =  matrix(beta1,lmax+1,nsim,byrow=FALSE)
truthmat2 =  matrix(beta2,lmax+1,nsim,byrow=FALSE)

mbhat1 = apply(bhat1,1,mean)
mbhat2 = apply(bhat2,1,mean)

mmboot = apply(mboot,1,mean)
mmboot2 = apply(mboot2,1,mean)


truthmat1 =  matrix( mbhat1,lmax+1,nsim,byrow=FALSE)
truthmat2 =  matrix( mbhat2,lmax+1,nsim,byrow=FALSE)



covnoboot = apply( (truthmat1 >= bhat1 - 2*sdnoboot) &  (truthmat1 <= bhat1 + 2*sdnoboot),1,mean)
covboot1 = apply( (truthmat1 >= bhat1 - 2*sdboot) &  (truthmat1 <= bhat1 + 2*sdboot),1,mean)
covboot.2 = apply( (truthmat1 > bhat1 - 2*sdboot.2) &  (truthmat1 < bhat1 + 2*sdboot.2),1,mean)

covquantboot1 = apply( (truthmat1 >= quantboot[1,,]) & (truthmat1 <= quantboot[2,,]),1,mean)
covquantboot.2 = apply( (truthmat1 >= quantboot.2[1,,]) & (truthmat1 <= quantboot.2[2,,]),1,mean)

covnoboot2 = apply( (truthmat2 >= bhat2 - 2*sdnoboot2) &  (truthmat2 <= bhat2 + 2*sdnoboot2),1,mean)
covboot2 = apply( (truthmat2 >= bhat2 - 2*sdboot2) &  (truthmat2 <= bhat2 + 2*sdboot2),1,mean)
covboot2.2 = apply( (truthmat2 > bhat2 - 2*sdboot2.2) &  (truthmat2 < bhat2 + 2*sdboot2.2),1,mean)

covquantboot2 = apply( (truthmat2 >= quantboot2[1,,]) & (truthmat2 <= quantboot2[2,,]),1,mean)
covquantboot2.2 = apply( (truthmat2 >= quantboot2.2[1,,]) & (truthmat2 <= quantboot2.2[2,,]),1,mean)


X11()
par(mfrow=c(2,1),mar=c(5,5,1,1))
matplot(cbind(covboot1,covboot.2),
        col=c(2,2),lty=c(1,2),type='l',ylab='Coverage Probability',xlab='lag',
        cex.lab=1.5,cex.axis=1.5,ylim=c(0.9,1),lwd=2)
abline(h=0.95)
matplot(cbind(covboot2,covboot2.2),
        col=c(2,2),lty=c(1,2),type='l',ylab='Coverage Probability',xlab='lag',
        cex.lab=1.5,cex.axis=1.5,ylim=c(0.9,1),lwd=2)
abline(h=0.95)


postscript('blah.eps',paper='special',width=10,height=6)
par(pin=c(8,4))
matplot(cbind(covboot1,covboot2,covnoboot,covnoboot2),type='l',lty=c(1,2,1,2),ylab='Coverage Probability',xlab='lag',
        cex.lab=1.5,cex.axis=1.5,ylim=c(0.5,1),lwd=2,col=c('black','black','darkgrey','darkgrey'))
abline(h=0.95)
legend('bottomleft',c('Bootstrap Velocity','Bootstrap Acceleration','Delta Velocity','Delta Acceleration'),lty=c(1,2,1,2),lwd=2,col=c('black','black','darkgrey','darkgrey'),cex=1.5)
dev.off()

save.image(savefile)
