# This file runs a simulation to test the statistical properties
# of functional convolution models. This file should be run first,
# it will save some needed material after which run ConvolutionSimCode2.r
#
# Most of this follows 1TruckAnalysis_final.r which is more thoroughly
# commented. For historical reasons, we have smoothed the covariates less
# here; for memory reasons it will be helpful to save some objects and throw
# the rest away.

library('fda')
library('Matrix')

savefile = 'TruckRun.Simdata2.Rdata'

runnames = c('Cycle 1','Cycle 2','Cycle 3','Cycle 4')

truck = read.table('1truck.dat',head=TRUE)

X = list()

which = c(4,6,2,5)
for(j in 1:length(which)){

  i = which[j]
  X[[j]] = matrix(truck[truck[,3]==i,1],sum(truck[,3]==i),1)

  rr = c(1,length(X[[j]]))
  bb = create.bspline.basis(rr,norder=6,breaks=1:rr[2])

  fd0 = smooth.basis(1:rr[2],X[[j]],fdPar(bb,3,1e-12))

  v = eval.fd(1:rr[2],fd0$fd,1)

  X[[j]] = cbind(X[[j]],v)
}

print('X')

lt = 20
lmax = 40
lmin = 5



ncyc = 4
n = rep(0,ncyc)

Xmat = list(len=ncyc)
YY = c()

nremove = c(9,19,18,9)

for(i in 1:ncyc){
  n[i] = nrow(X[[i]])
  Xmat[[i]] = list(len=ncol(X[[i]]))
  for(j in 1:ncol(X[[i]])){
    Xmat[[i]][[j]] =  matrix(X[[i]][,j],nrow(X[[i]])+1,(lmax+1),byrow=FALSE)[1:(n[i]-lmax-1),rev(1:(lmax+1))]
    Xmat[[i]][[j]] = Xmat[[i]][[j]][-(1:nremove[i]),]
  }
  n[i] = nrow(Xmat[[i]][[1]])
}

N = cumsum(c(1,n))


print('Xmat')

# Now put this into bases

basislist = vector('list',len=lmax-lmin)
basismatlist = vector('list',len=lmax-lmin)
Xmatlist = vector('list',len=lmax-lmin)
Penmatlist =  vector('list',len=lmax-lmin)
b = rep(0,lmax-lmin)


for(i in 1:(lmax-lmin)){
  basislist[[i]] = create.bspline.basis(c(0,i+lmin),norder=4,breaks = 0:(i+lmin))
  b[i] = basislist[[i]]$nbasis

  tPen = eval.penalty(basislist[[i]],int2Lfd(2))
  Penmatlist[[i]] =  rbind(matrix(0,4,4+2*(ncol(tPen)-1)),cbind(matrix(0,2*(nrow(tPen)-1),4),diag(2)%x%tPen[-b[i],-b[i]]))
#  Penmatlist[[i]] =  rbind(rep(0,1+1*(ncol(tPen)-1)),cbind(rep(0,1*(nrow(tPen)-1)),diag(1)%x%tPen[-b[i],-b[i]]))
  basismatlist[[i]] = eval.basis(0:(i+lmin),basislist[[i]])[,-b[i]]

  for(j in 1:ncyc){
    tX = matrix(0,n[j],ncyc)
    tX[,j] = 1
    for(k in 1:ncol(X[[j]])){
#      tX = cbind(tX,Xmat[[j]][[k]][,lmax+1-(0:(i+lmin))]%*%basismatlist[[i]])
      tX = cbind(tX,Xmat[[j]][[k]][,1+(0:(i+lmin))]%*%basismatlist[[i]])
    }
    Xmatlist[[i]] = rbind(Xmatlist[[i]],tX)
  }
}

print('Xmatlist')

# Hat matrices

lambda = 10^seq(-3,1,0.25)

hatmat = vector('list',len=length(lambda))

df = matrix(0,length(lambda),lmax-lmin)

for(i in 1:length(lambda)){
   hatmat[[i]] = vector('list',len=lmax-lmin)
   for(j in 1:(lmax-lmin)){
     hatmat[[i]][[j]] = solve(t(Xmatlist[[j]])%*%Xmatlist[[j]]+lambda[i]*Penmatlist[[j]],t(Xmatlist[[j]]))
     df[i,j] = sum(diag(Xmatlist[[j]]%*%hatmat[[i]][[j]]))
   }
}


print('hatmat')

nboot = 500
I = matrix(1:3140,20,157)
K = 20
M = 157

betas = read.table('truebetas.dat',sep=',')

beta1 = betas[,1]
beta2 = betas[,2]

save(basislist,basismatlist,beta1,beta2,hatmat,I,K,lambda,lmax,lmin,lt,M,n,N,ncyc,nboot,Xmat,Xmatlist,file=savefile)


